"""
.. deprecated:: 0.20.1
   Use :py:mod:`armarx_core.ice_manager` instead.
"""

import warnings
from armarx_core.ice_manager import *

warnings.warn("Use armarx_core.ice_manager instead", DeprecationWarning)
